/*
 * Copyright 2019 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.kafka.secretregistry.crypto;

import static java.nio.charset.StandardCharsets.UTF_8;
import static javax.crypto.Cipher.DECRYPT_MODE;
import static javax.crypto.Cipher.ENCRYPT_MODE;

import com.google.common.io.BaseEncoding;

import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.Provider;
import java.security.SecureRandom;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.Mac;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKey;
import javax.crypto.spec.GCMParameterSpec;
import javax.crypto.spec.SecretKeySpec;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Cryptor which encrypts/decrypts secret content, based on KeyWhiz.
 *
 * <p>Encryption keys are derived using a provided info tag.
 */
public class Cryptor {
  private static final Logger logger = LoggerFactory.getLogger(Cryptor.class);

  private static final String ENCRYPTION_ALGORITHM = "AES/GCM/NoPadding";
  private static final String KEY_ALGORITHM = "AES";
  private static final int TAG_BITS = 128;
  private static final int NONCE_BYTES = 12;

  private final SecretKey key;
  private final Provider provider;
  private final SecureRandom random;

  public Cryptor(SecretKey key, Provider provider, SecureRandom random) {
    this.key = key;
    this.provider = provider;
    this.random = random;
  }

  public class Encrypter {
    private final String derivationInfo;

    private Encrypter(String derivationInfo) {
      this.derivationInfo = derivationInfo;
    }

    /**
     * Encrypts content under a derived key.
     *
     * @param plaintext plaintext in bytes
     * @return encrypted object containing ciphertext and parameters necessary for decryption
     */
    public Encrypted encrypt(byte[] plaintext) {
      byte[] nonce = new byte[NONCE_BYTES];
      random.nextBytes(nonce);

      byte[] ciphertext = gcm(Mode.ENCRYPT, derivationInfo, nonce, plaintext);
      return Encrypted.of(derivationInfo, ciphertext, nonce);
    }
  }

  /**
   * Builds an encrypter using key derived from the provided secret name.
   *
   * @param secretName non-empty secret name used for key derivation
   * @return encrypter capable of operating on plaintext data
   */
  public Encrypter encryptionKeyDerivedFrom(String secretName) {
    return new Encrypter(secretName);
  }

  /**
   * Decrypts content previously encrypted by {@link Cryptor}.
   *
   * @param encrypted object containing ciphertext from prior {@link Encrypter#encrypt} call
   * @return original plaintext in bytes
   */
  public byte[] decrypt(Encrypted encrypted) {
    return gcm(Mode.DECRYPT, encrypted.derivationInfo(),
        encrypted.ivBytes(), encrypted.contentBytes());
  }

  public String computeHmac(byte[] data) {
    SecretKey hmacKey = deriveKey(32, "hmackey");
    try {
      Mac mac = Mac.getInstance("HmacSHA256");
      mac.init(hmacKey);
      return BaseEncoding.base16().encode(mac.doFinal(data));
    } catch (NoSuchAlgorithmException | InvalidKeyException e) {
      logger.warn("Error computing HMAC: ", e);
      return null;
    }
  }

  private SecretKey deriveKey(int blockSize, String info) {
    Hkdf hkdf = Hkdf.usingProvider(provider);
    byte[] infoBytes = info.getBytes(UTF_8);
    byte[] derivedKeyBytes = hkdf.expand(key, infoBytes, blockSize);
    return new SecretKeySpec(derivedKeyBytes, KEY_ALGORITHM);
  }

  private byte[] gcm(Mode mode, String info, byte[] nonce, byte[] data) {
    try {
      Cipher cipher = Cipher.getInstance(ENCRYPTION_ALGORITHM, provider);
      SecretKey derivedKey = deriveKey(cipher.getBlockSize(), info);
      GCMParameterSpec gcmParameters = new GCMParameterSpec(TAG_BITS, nonce);
      cipher.init(mode.cipherMode, derivedKey, gcmParameters);
      return cipher.doFinal(data);
    } catch (IllegalBlockSizeException | InvalidAlgorithmParameterException
        | NoSuchPaddingException | NoSuchAlgorithmException
        | InvalidKeyException | BadPaddingException e) {
      throw new RuntimeException(e);
    }
  }

  private enum Mode {
    ENCRYPT(ENCRYPT_MODE), DECRYPT(DECRYPT_MODE);

    final int cipherMode;

    Mode(int cipherMode) {
      this.cipherMode = cipherMode;
    }
  }
}
