/*
 * Copyright 2019 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.kafka.secretregistry.rest.resources;

import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.validation.Valid;
import javax.ws.rs.Consumes;
import javax.ws.rs.DELETE;
import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.container.AsyncResponse;
import javax.ws.rs.container.Suspended;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.HttpHeaders;
import javax.ws.rs.core.SecurityContext;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import io.confluent.kafka.secretregistry.client.rest.Versions;
import io.confluent.kafka.secretregistry.client.rest.entities.Secret;
import io.confluent.kafka.secretregistry.exceptions.SecretRegistryException;
import io.confluent.kafka.secretregistry.exceptions.SecretRegistryStoreException;
import io.confluent.kafka.secretregistry.rest.exceptions.Errors;
import io.confluent.kafka.secretregistry.storage.KafkaSecretRegistry;
import io.confluent.rest.annotations.PerformanceMetric;

@Path("/secret/paths")
@Produces({Versions.SECRET_REGISTRY_V1_JSON_WEIGHTED,
    Versions.SECRET_REGISTRY_DEFAULT_JSON_WEIGHTED,
    Versions.JSON_WEIGHTED})
@Consumes({Versions.SECRET_REGISTRY_V1_JSON,
    Versions.SECRET_REGISTRY_DEFAULT_JSON,
    Versions.JSON, Versions.GENERIC_REQUEST})
public class PathResource {

  private static final Logger log = LoggerFactory.getLogger(PathResource.class);
  private final KafkaSecretRegistry secretRegistry;
  private final RequestHeaderBuilder requestHeaderBuilder = new RequestHeaderBuilder();

  public PathResource(KafkaSecretRegistry secretRegistry) {
    this.secretRegistry = secretRegistry;
  }


  @GET
  @Valid
  @PerformanceMetric("paths.list")
  public Set<String> list(
      @Context SecurityContext securityContext
  ) {
    if (securityContext.getUserPrincipal() != null) {
      String name = securityContext.getUserPrincipal().getName();
      log.info("User '{}' listing all paths", name);
    } else {
      log.info("Unknown user listing all paths");
    }

    try {
      return secretRegistry.listPaths();
    } catch (SecretRegistryStoreException e) {
      throw Errors.storeException("Error while listing paths", e);
    } catch (SecretRegistryException e) {
      throw Errors.secretRegistryException("Error while listing paths", e);
    }
  }

  @GET
  @Path("/{path}")
  @PerformanceMetric("paths.get-latest")
  public List<Secret> allLatest(
      @Context SecurityContext securityContext,
      @PathParam("path") String path
  ) {
    if (securityContext.getUserPrincipal() != null) {
      String name = securityContext.getUserPrincipal().getName();
      log.info("User '{}' getting latest for '{}'", name, path);
    } else {
      log.info("Unknown user getting latest for '{}'", path);
    }

    try {
      if (!secretRegistry.listPaths().contains(path)) {
        throw Errors.pathNotFoundException();
      }
      return secretRegistry.getAllLatest(path);
    } catch (SecretRegistryStoreException e) {
      throw Errors.storeException("Error while listing paths", e);
    } catch (SecretRegistryException e) {
      throw Errors.secretRegistryException("Error while listing paths", e);
    }
  }

  @DELETE
  @Path("/{path}")
  @PerformanceMetric("paths.delete-path")
  public void deletePath(
      final @Suspended AsyncResponse asyncResponse,
      @Context SecurityContext securityContext,
      @Context HttpHeaders headers,
      @PathParam("path") String path
  ) {
    if (securityContext.getUserPrincipal() != null) {
      String name = securityContext.getUserPrincipal().getName();
      log.info("User '{}' deleting path '{}'", name, path);
    } else {
      log.info("Unknown user deleting path '{}'", path);
    }

    Set<String> deletedKeys;
    try {
      if (!secretRegistry.listPaths().contains(path)) {
        throw Errors.pathNotFoundException();
      }
      Map<String, String> headerProperties = requestHeaderBuilder.buildRequestHeaders(
          headers, secretRegistry.config().whitelistHeaders());
      deletedKeys = secretRegistry.deletePathOrForward(headerProperties, path);
    } catch (SecretRegistryException e) {
      throw Errors.secretRegistryException(
          "Error while deleting the path " + path,
          e
      );
    }
    asyncResponse.resume(deletedKeys);
  }

}
